#include <torch/extension.h>
#include <ATen/ATen.h>
#include <cuda_runtime.h>
#include <vector>
#include <string>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

namespace py = pybind11;

// Warp-level reduction using __shfl_down_sync
__inline__ __device__ float warpReduceSum(float val) {
    for (int offset = warpSize/2; offset > 0; offset /= 2) {
        val += __shfl_down_sync(0xffffffff, val, offset);
    }
    return val;
}

// Optimized CUDA kernel with strided loops and efficient work distribution
__global__ void globalAvgPoolKernel(const float* __restrict__ input,
                                   float* __restrict__ output,
                                   const int N,
                                   const int C,
                                   const int H,
                                   const int W) {
    const int HW = H * W;
    const int tid = threadIdx.x;
    const int bid = blockIdx.x;
    const int n = bid / C;
    const int c = bid % C;
    
    // Base offset for this feature map
    const int base_idx = (n * C + c) * HW;
    
    // Compute optimal stride based on thread count and workload
    const int stride = blockDim.x * 4;  // Process 4 elements per thread per iteration
    float sum = 0.0f;
    
    // Main loop with stride optimization
    #pragma unroll 4
    for (int i = tid * 4; i < HW; i += stride) {
        // Load 4 elements per iteration if possible
        if (i + 3 < HW) {
            sum += input[base_idx + i] +
                   input[base_idx + i + 1] +
                   input[base_idx + i + 2] +
                   input[base_idx + i + 3];
        } else {
            // Handle boundary case
            for (int j = 0; j < 4 && (i + j) < HW; j++) {
                sum += input[base_idx + i + j];
            }
        }
    }
    
    // Warp-level reduction
    sum = warpReduceSum(sum);
    
    // Block-level reduction using shared memory
    __shared__ float shared_sum;
    if (tid == 0) {
        shared_sum = 0.0f;
    }
    __syncthreads();
    
    if (tid == 0) {
        atomicAdd(&shared_sum, sum);
    }
    __syncthreads();
    
    // Write final result
    if (tid == 0) {
        output[bid] = shared_sum / static_cast<float>(HW);
    }
}

// Host function for global average pooling
at::Tensor global_avg_pool(at::Tensor x) {
    auto input = x.contiguous();
    const int N = input.size(0);
    const int C = input.size(1);
    const int H = input.size(2);
    const int W = input.size(3);
    
    at::Tensor output = at::empty({N, C}, input.options());
    
    const int threads = 128;  // Increased thread count for better occupancy
    const int blocks = N * C;
    
    globalAvgPoolKernel<<<blocks, threads>>>(
        input.data_ptr<float>(),
        output.data_ptr<float>(),
        N, C, H, W
    );
    
    return output;
}

// Dense layer implementation
at::Tensor dense_layer_fn(
    at::Tensor x,
    at::Tensor bn_weight,
    at::Tensor bn_bias,
    at::Tensor bn_mean,
    at::Tensor bn_var,
    at::Tensor conv_weight,
    bool is_training
) {
    x = at::batch_norm(
        x, bn_weight, bn_bias, bn_mean, bn_var,
        is_training, 0.1, 1e-5, true
    );
    x = at::relu(x);
    x = at::conv2d(x, conv_weight, /*bias=*/{}, /*stride=*/{1, 1}, /*padding=*/{1, 1});
    return x;  // Removed dropout as p=0.0
}

// Transition layer implementation
at::Tensor transition_layer_fn(
    at::Tensor x,
    at::Tensor bn_weight,
    at::Tensor bn_bias,
    at::Tensor bn_mean,
    at::Tensor bn_var,
    at::Tensor conv_weight,
    bool is_training
) {
    x = at::batch_norm(
        x, bn_weight, bn_bias, bn_mean, bn_var,
        is_training, 0.1, 1e-5, true
    );
    x = at::relu(x);
    x = at::conv2d(x, conv_weight);
    x = at::avg_pool2d(x, /*kernel_size=*/{2, 2}, /*stride=*/{2, 2});
    return x;
}

// Main module function
at::Tensor module_fn(
    at::Tensor x,
    py::object params,
    bool is_training
) {
    auto get_param = [&](const std::string& key) -> at::Tensor {
        return params.attr("__getitem__")(key.c_str()).cast<at::Tensor>();
    };

    // Initial features
    x = at::conv2d(x, get_param("features_conv_weight"), 
                   /*bias=*/{}, /*stride=*/{2, 2}, /*padding=*/{3, 3});
    
    x = at::batch_norm(
        x, 
        get_param("features_bn_weight"),
        get_param("features_bn_bias"),
        get_param("features_bn_mean"),
        get_param("features_bn_var"),
        is_training, 0.1, 1e-5, true
    );
    x = at::relu(x);
    x = at::max_pool2d(x, /*kernel_size=*/{3, 3}, /*stride=*/{2, 2}, /*padding=*/{1, 1});

    std::vector<int> num_layers = {6, 12, 24, 16};

    // Dense blocks and transitions
    for (int i = 0; i < 4; ++i) {
        std::vector<at::Tensor> features;
        features.push_back(x);

        for (int j = 0; j < num_layers[i]; ++j) {
            std::string prefix = "block" + std::to_string(i) + "_layer" + std::to_string(j) + "_";
            
            at::Tensor new_feature = dense_layer_fn(
                x,
                get_param(prefix + "bn_weight"),
                get_param(prefix + "bn_bias"),
                get_param(prefix + "bn_mean"),
                get_param(prefix + "bn_var"),
                get_param(prefix + "conv_weight"),
                is_training
            );

            features.push_back(new_feature);
            x = at::cat(features, 1);
        }

        if (i != 3) {
            std::string prefix = "transition" + std::to_string(i) + "_";
            x = transition_layer_fn(
                x,
                get_param(prefix + "bn_weight"),
                get_param(prefix + "bn_bias"),
                get_param(prefix + "bn_mean"),
                get_param(prefix + "bn_var"),
                get_param(prefix + "conv_weight"),
                is_training
            );
        }
    }

    // Final layers
    x = at::batch_norm(
        x,
        get_param("final_bn_weight"),
        get_param("final_bn_bias"),
        get_param("final_bn_mean"),
        get_param("final_bn_var"),
        is_training, 0.1, 1e-5, true
    );
    x = at::relu(x);
    
    if (x.is_cuda()) {
        x = global_avg_pool(x).view({x.size(0), -1});
    } else {
        x = at::adaptive_avg_pool2d(x, {1, 1}).reshape({x.size(0), -1});
    }

    x = at::linear(x, get_param("classifier_weight"), get_param("classifier_bias"));
    return x;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_fn, "DenseNet121 forward with optimized strided loops");
}